"""
Cloud2Sql
~~~~~~~~~~~~~~~~~~~~~~~~~~
Read infrastructure data from your cloud and export it to a SQL database.
:copyright: © 2022 Some Engineering Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

__title__ = "cloud2sql"
__description__ = "Read infrastructure data from your cloud and export it to a SQL database."
__author__ = "Some Engineering Inc."
__license__ = "Apache 2.0"
__copyright__ = "Copyright © 2022 Some Engineering Inc."
__version__ = "0.6.0"
